/* $Id: gdbm_table.c,v 1.2 2001/11/25 15:23:40 aa5779 Exp $ */
/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <gdbm.h>
#include "util.h"
#include "dstring.h"
#include "lists.h"
#include "interp.h"
#define TABLE_IMP
#include "table.h"


static void error_db(char *msg)
{
	recover_error("GDBM error: %s", msg);
}

storage_class table_storage_class = {table_create, table_destroy, 
									 table_check, table_get,
									 table_put, table_add,
									 table_remove,
									 table_first, table_next,
									 table_enditer,
									 table_numofitems
									};

typedef GDBM_FILE db_t;

typedef datum dbdata_t;

#define db_error() (gdbm_strerror(gdbm_errno))

DEFLIST(dbf_list, 
		db_t db;
		);

static int is_always_new;
static int read_only;


static void *open_db(char *fname)
{
	return gdbm_open(fname, 2048, is_always_new ? GDBM_NEWDB : (read_only ? GDBM_READER : GDBM_WRCREAT), S_IREAD | S_IWRITE, error_db);
}
	
static dbf_list *all_tables;

#define TABLE_MAX_VAL  "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF" \
					   "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"

storage *table_create(interpreter *i, char *name, char *arg)
{
	storage *ns = ALLOC(storage);
	db_t db;
	dbf_list *ndbl;
	dbdata_t key = {"\0\0\0\0", 4};
	dbdata_t maxval = {TABLE_MAX_VAL, 17
				};
	int dummy = 0;
	char *primary_env = NULL;

   	is_always_new = (*arg == '!');
	read_only = (*arg == '+');
	arg += is_always_new || read_only;
	if(is_always_new)
		primary_env = getenv(SL_TABLE_NEW_PATH_ENV);
	else if(read_only)
		primary_env = getenv(SL_TABLE_RO_PATH_ENV);
	if(!primary_env)
		primary_env = getenv(SL_TABLE_PATH_ENV);
	db = find_filename(arg, primary_env, "t", open_db, open_db);
	if(!db)
		recover_error("can't load DB '%s'(%s): %s", name, arg, db_error());
	ns->data = db;
	ns->class = &table_storage_class;
	ndbl = ALLOC(dbf_list);
	ndbl->db = db;
	all_tables = list_add(all_tables, ndbl);
	if(!gdbm_exists(db, key))
	{
		gdbm_store(db, key, maxval, GDBM_INSERT);
		key.dptr = "\0\0reccount";
		key.dsize = 10;
		maxval.dptr = (char *)&dummy;
		maxval.dsize = sizeof(dummy);
		gdbm_store(db, key, maxval, GDBM_INSERT);
		db->put(db, &key, &maxval, 0);
	}
	return ns;
}

static int find_table_in_list(dbf_list *w, db_t key)
{
	return w->db == key;
}

extern int si_find_byso(shared_int *si, void *so);
static char *get_table_name(interpreter *i, storage *s)
{
	shared_int *si;
	if(!i)
		return "<unknown>";
	si = list_find(i->stos, s, (list_predicate_t)si_find_byso);
	return si ? si->intname : "<unknown>";
}

static void destroy_db(dbf_list *w)
{
    gdbm_close(w->db);
}

void table_destroy(interpreter *i, storage *s)
{
	db_t db = s->data;
	all_tables = list_remove(all_tables, db, (list_predicate_t)find_table_in_list,
				(list_destroy_t)destroy_db, sizeof(dbf_list)
				);
}

int table_check(interpreter *i, storage *s, dstring *key)
{
	dbdata_t ky;
	ky.dsize = ds_length(key);
	ky.dptr = DS_BODY(key);
	return gdbm_exists(s->data, ky);
}

extern dstring *ds_icreate(char *data, int len);

dstring *table_get(interpreter *i, storage *s, dstring *key)
{
	dbdata_t ky, data;
	
	ky.dsize = ds_length(key);
	ky.dptr = DS_BODY(key);
	data = gdbm_fetch(s->data, ky);
	if(!data.dptr)
		recover_error("No '%s' in table '%s'", DS_BODY(key), get_table_name(i, s));
	return ds_icreate(data.dptr, data.dsize);
}

void table_put(interpreter *i, storage *s, dstring *key, dstring *val)
{
	dbdata_t ky, tostore;
	
	ky.dsize = ds_length(key);
	ky.dptr  = DS_BODY(key);
	if(!gdbm_exists(s->data, ky))
		recover_error("No '%s' in table '%s'", DS_BODY(key), get_table_name(i, s));
	tostore.dsize = ds_length(val) + 1;
	tostore.dptr = DS_BODY(val);
	gdbm_store(s->data, ky, tostore, GDBM_REPLACE);
}

static dbdata_t find_prev_rec(db_t db, char *key, dbdata_t *d)
{
	dbdata_t ky, val = {NULL, 0};
	ky.dsize = 4;
	ky.dptr = malloc(4);
	memset(ky.dptr, 0, 4);
	for(;;)
	{
		val = gdbm_fetch(db, ky);
		if(strcmp(val.dptr, key) > 0)
			break;
		free(ky.dptr);
		ky = val;
	}
	*d = val;
	return ky;
}

void table_add(interpreter *i, storage *s, dstring *key, dstring *val)
{
	dbdata_t vl;
	dbdata_t prevk;
	dbdata_t ky;
	dbdata_t tostore;
	
	ky.dsize = ds_length(key);
	ky.dptr = DS_BODY(key);
	tostore.dsize = ds_length(val) + 1;
	tostore.dptr = DS_BODY(val);
	if(gdbm_store(s->data, ky, tostore, GDBM_INSERT))
		recover_error("'%s' already in table '%s'", DS_BODY(key), get_table_name(i, s));
	prevk = find_prev_rec(s->data, DS_BODY(key), &vl);
	ky.dsize++;
	gdbm_store(s->data, ky, vl, GDBM_INSERT);
	free(vl.dptr);
	gdbm_store(s->data, prevk, ky, GDBM_REPLACE);
	free(prevk.dptr);
	ky.dsize = 10;
	ky.dptr = "\0\0reccount";
	vl = gdbm_fetch(s->data, ky);
	(*(int *)vl.dptr)++;
	gdbm_store(s->data, ky, vl, GDBM_REPLACE);
	free(vl.dptr);
}

static dbdata_t find_prev_rec1(db_t db, char *key)
{
	dbdata_t ky, val = {NULL, 0};
	ky.dsize = 4;
	ky.dptr = malloc(4);
	memset(ky.dptr, 0, 4);
	for(;;)
	{
		val = gdbm_fetch(db, ky);
		if(!strcmp(val.dptr, key))
			break;
		free(ky.dptr);
		ky = val;
	}
	free(val.dptr);
	return ky;
}

void table_remove(interpreter *i, storage *s, dstring *key)
{
	dbdata_t ky;
	dbdata_t val;
	ky.dsize = ds_length(key);
	ky.dptr = DS_BODY(key);
	if(gdbm_delete(s->data, ky))
		recover_error("'%s' not in table '%s'", DS_BODY(key), get_table_name(i, s));
	ky.dsize++;
	val = gdbm_fetch(s->data, ky);
	gdbm_delete(s->data, ky);
	ky = find_prev_rec1(s->data, DS_BODY(key));
	gdbm_store(s->data, ky, val, GDBM_REPLACE);
	free(val.dptr);
	free(ky.dptr);
	ky.dsize = 10;
	ky.dptr = "\0\0reccount";
	val = gdbm_fetch(s->data, ky);
	(*(int *)val.dptr)--;
	gdbm_store(s->data, ky, val, GDBM_REPLACE);
	free(val.dptr);
}

dstring *table_first(interpreter *i, storage *s, void **iter)
{
	dbdata_t ky;
	dbdata_t val;
	ky.dsize = 4;
	ky.dptr = "\0\0\0\0";
	val = gdbm_fetch(s->data, ky);
	if(!strcmp(val.dptr, TABLE_MAX_VAL))
	{
		free(val.dptr);
		return NULL;
	}
	*iter = ds_fix(ds_icreate(val.dptr, val.dsize));
	return *iter;
}

dstring *table_next(interpreter *i, storage *s, void **iter)
{
	dbdata_t ky, val;
	ky.dsize = ds_length(*iter) + 1;
	ky.dptr = DS_BODY((dstring *)*iter);
	val = gdbm_fetch(s->data, ky);
	ds_unfix(*iter);
	if(!strcmp(val.dptr, TABLE_MAX_VAL))
		return NULL;
	*iter = ds_fix(ds_icreate(val.dptr, val.dsize));
	return *iter;
}

void table_enditer(interpreter *i, storage *s, void **iter)
{
	if(*iter)
		ds_unfix(*iter);
	*iter = NULL;
}

int table_numofitems(interpreter *i, storage *s)
{
	dbdata_t ky, val;
	int c;
	
	ky.dsize = 10;
	ky.dptr = "\0\0reccount";
	val = gdbm_fetch(s->data, ky);
	c = *(int *)val.dptr;
	free(val.dptr);
	return c;
}

