/* $Id: sl.c,v 1.23 2002/07/15 22:07:07 aa5779 Exp $ */
/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/// \section[COMMAND LINE OPTIONS]
// \em[-i infile] -- name of the standard input stream file (none by default)
// 
// \em[-o outfile] -- name of the standard output stream file (console by
// default); \em[none] means no output stream opened; the same effect may be
// achieved by using \em[-nout] as a part of the program name.
//
// \em[-d preload-modules] -- name of an alternative file containing the
// list preloaded modules (by default, preload.scm in \em2[STD_LOCATION][)]
//
// \em[-l location] -- path to alternative location for NSL system files,
// (\em[STD_LOCATION] by default)
//
// \em[-s script-path] -- a path added to a path list for scripts
//
// \em[-c] -- run sl as a CGI client. The same effect may be achieved by
// using \em[cgi] as a part of the program name
//
// \em[-f] -- using flat layout, that is, scripts and automata are considered to be
// at the top of the standard location (useful for running without installation)
//
// \em[-g squeeze-interval] -- the number of cycles between squeezing unused strings
//
// \em[-G squeeze-long-string] -- the size of a string in Kbytes considered long, so that
// its memory is really freed when squeezing
// 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "util.h"
#include "dstring.h"
#include "interp.h"
#include "bstream.h"
#include "cgistream.h"
#ifdef USE_LTDL
#include "modules.h"
#endif
#include "trap.h"

extern int program_argc;
extern char **program_argv;
extern int squee_interval;
extern int squee_long_string;
char *nsl_location = NULL;

static stream *outs, *ins;
static int no_out;
static int as_cgi;
static int flat_layout;
static char *def_mod;
static char *location = STD_LOCATION;


static char *add_script_path;
// static int from_cur_dir;

static void usage(char *argv0)
{
	fprintf(stderr, "Usage: %s [-fc] [-l location] [-s script-path] [-d preload-file] [-g squeeze-interva] [-G squeeze-long-string] [-i infile] [-o outfile] script-file [arguments ...]\n", argv0);
}

static void parse_options(int argc, char *argv[])
{
	int c;
	opterr = 0;
	if(strstr(argv[0], "cgi"))
		as_cgi = 1;
	if(strstr(argv[0], "-nout"))
		no_out = 1;
	for(;;)
	{
		c = getopt(argc, argv, "i:o:hucg:G:d:l:s:f");
		switch(c)
		{
			case -1:
				if(!argv[optind])
				{
					fputs("No script file name given\n", stderr);
					usage(*argv);
					exit(EXIT_FAILURE);
				}
				program_argv = argv + optind;
				program_argc = argc - optind;
				return;
			case 'c':
				if(outs || ins)
					fatal_error("can't specify input/output streams with '-c' switch");
				as_cgi = 1;
				break;
			case '?':
				fprintf(stderr, "Unrecognized option character: '%c'\n", optopt);
			case 'h':
			case 'u':
				usage(*argv);
				exit(EXIT_FAILURE);
			case 'i':
				if(as_cgi)
					fatal_error("can't specify input/output streams with '-c' switch");
				ins = basic_stream_class.open(NULL, "stdin", *optarg == '-' && !optarg[1] ? "" : optarg, 0);
				break;
			case 'o':
				if(!strcmp(optarg, "none"))
					no_out = 1;
				else
				{
					if(as_cgi)
						fatal_error("can't specify input/output streams with '-c' switch");	
					else outs = basic_stream_class.open(NULL, "stdout", *optarg == '-' && !optarg[1] ? "" : optarg, 1);
				}
				break;
			case 'g':
				squee_interval = strtoul(optarg, NULL, 10);
				break;
			case 'G':
				squee_long_string = strtoul(optarg, NULL, 10) * 1000;
				break;
			case 'd':
				def_mod = optarg;
				break;
			case 'l':
				location = optarg;
				break;
			case 's':
				add_script_path = optarg;
				break;
			case 'f':
				flat_layout = 1;
				break;
		}
	}
}

static void setup_locations(void)
{
	dstring *temp = ds_create(location);
	dstring *temp1 = ds_create(NULL);
	if(DS_BODY(temp)[ds_length(temp) - 1] != '/')
		ds_appendch(temp, '/');
	nsl_location = xstrdup(DS_BODY(temp));
	if(flat_layout)
	{
		setenv(SL_TABLE_RO_PATH_ENV, DS_BODY(temp), 0);
		setenv(AU_PATH_ENV, DS_BODY(temp), 0);
		if(as_cgi)
			ds_appendstr(temp1, "./:");
		if(add_script_path)
		{
			ds_appendstr(temp1, add_script_path);
			ds_appendch(temp1, ':');
		}
		ds_append(temp1, temp);
		setenv(SCRIPT_PATH_ENV, DS_BODY(temp1), 0);
	}
	else
	{
		if(as_cgi)
			ds_appendstr(temp1, "./:");
		if(add_script_path)
		{
			ds_appendstr(temp1, add_script_path);
			ds_appendch(temp1, ':');
		}
		ds_append(temp1, temp);
		ds_appendstr(temp1, "scripts:");
		ds_append(temp1, temp);
		ds_appendstr(temp1, "test");
		ds_appendch(temp1, ':');
		if(getenv("PATH"))
			ds_appendstr(temp1, getenv("PATH"));
		setenv(SCRIPT_PATH_ENV, DS_BODY(temp1), 0);
		ds_assign(temp1, temp);
		ds_appendstr(temp1, "tables");
		setenv(SL_TABLE_RO_PATH_ENV, DS_BODY(temp1), 0);
		ds_assign(temp1, temp);
		ds_appendstr(temp1, "automata:");
		ds_append(temp1, temp);
		ds_appendstr(temp1, "test");
		setenv(AU_PATH_ENV, DS_BODY(temp1), 0);
	}
}

int main(int argc, char *argv[])
{
	int exit_code;
	interpreter *i;
	shared_int *toplevel;
	
	register_error_handler(NULL);
	if(!CATCH())
	{
		parse_options(argc, argv);
		setup_locations();
		if(!as_cgi)
		{
			if(!outs && !no_out)
				outs = basic_stream_class.open(NULL, "stdout", "", 1);
		}
		else
		{
			ins = incgi_stream_class.open(NULL, "stdin", "", 0);
			if(!no_out)
				outs = outcgi_stream_class.open(NULL, "stdout", "", 1);
		}
	}
	unwind_error();
#ifdef USE_LTDL
	init_modules(def_mod);
#endif
	i = load_interpreter(*program_argv, ins, outs);
	toplevel = init_shared_object(i, i, "subint", "");
	export_shared(toplevel->so, "toplevel");
	register_error_handler(NULL);
	if(CATCH())
	{
		if(!as_cgi)
			print_error_msg("run-time error: %s", get_error_desc());
		else
		{
			puts("Content-Type: text/plain\n");
			fprintf(stdout, "SL/CGI (%s): Run-time error: %s at %s:%d\n", *program_argv, get_error_desc(), 
					set_current_filename(NULL), set_line_no(-1));
		}
		return 1;
	}
   	exit_code = script_main_loop(i);	
	cleanup_interpreter(i);
	return exit_code;
}
